<?php
namespace Service\DataWarehouse;

use Service\Manager\Mongo\Query\Expr;
use Service\Manager\Mongo\Types\DateType;
use Service\Manager\Mongo\Types\Type;
use Util\Utilities;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Zeedhi\Framework\DB\Mongo\Mongo;

class Manager implements ManagerInterface {

    /** @var Mongo */
    protected $mongoDB;
    protected $dateTimeFormat = "d/m/Y H:i:s";
    const DATE_COLUMN = 'DATA';

    /**
     * @param Mongo $mongoDB
     */
    function __construct(Mongo $mongoDB) {
        $this->mongoDB = $mongoDB;
    }

    /**
     * @param $row
     *
     * @return mixed
     */
    protected function normalizeRow($row) {
        foreach ($row as $key => $value) {
            if ($this->isDateTime($value) || $key === self::DATE_COLUMN) {
                $value = is_numeric($value) ? $value : $this->makeDateTime($value);
                $row[$key] = $this->parseDateToDatabaseValue($value);
            } else {
                $row[$key] = Type::convertPHPToDatabaseValue($value);
            }
        }
        return $row;
    }

    /**
     * @param $fact
     * @param $row
     *
     * @return bool
     * @throws Exception
     */
    public function addFact($fact, $row) {
        try {
            $row = $this->normalizeRow($row);
            $result = $this->mongoDB->insert($fact, $row);
            return $result['ok'] == 1;
        } catch (\Exception $e) {
            throw Exception::errorInAddFact($e);
        }
    }

    /**
     * @param $dateTimeFormat
     */
    public function setDateTimeFormat($dateTimeFormat) {
        $this->dateTimeFormat = $dateTimeFormat;
    }

    /**
     * @param $value
     *
     * @return bool
     */
    protected function isDateTime($value) {
        return $value instanceof \DateTime;
    }

    protected function makeDateTime($value) {
        return $this->isDateTime($value) ? $value : \DateTime::createFromFormat($this->dateTimeFormat, $value, DateType::getUTC());
    }

    protected function parseDateToDatabaseValue($value) {
        return Type::getType('date')->convertToDatabaseValue($value);
    }

    /**
     * @param FilterCriteria $criteria
     * @param                $start
     * @param                $end
     */
    protected function addDatesToFilter(FilterCriteria $criteria, $start, $end) {
        $start = $this->parseDateToDatabaseValue($this->makeDateTime($start));
        $end = $this->parseDateToDatabaseValue($this->makeDateTime($end));
        $criteria->addCondition(self::DATE_COLUMN, FilterCriteria::BETWEEN, array($start, $end));
    }

    protected function planningDocument($document) {
        foreach ($document as $column => $value) {
            $value = Type::convertDatabaseValueToPHP($value);
            if ($this->isDateTime($value)) {
                /** @var \DateTime $value */
                $value = $value->format($this->dateTimeFormat);
            }
            $document[$column] = $value;
        }
        return $document;
    }

    protected function prepareRowsForResultSet($rows) {
        $preparedRows = array();
        foreach ($rows as $row) {
            $preparedRows[] = $this->planningDocument($row);
        }
        return $preparedRows;
    }


    /**
     * @param FilterCriteria $criteria
     * @param Expr           $expressionBuilder
     */
    protected function processOrderBy(FilterCriteria $criteria, Expr $expressionBuilder) {
        if (count($order = $criteria->getOrderBy()) > 0) {
            $expressionBuilder->sort($order);
        }
    }

    protected function getTempCollectionName() {
        return 'temp_'.substr(uniqid(rand()), 0, 10);
    }

	/**
	 * @param $aggregateColumns
	 *
	 * @return array
	 */
	protected function buildAggregateExpr($aggregateColumns) {
		$pipeline = array(
			'$group' => array(
				'_id' => array()
			)
		);
		foreach ($aggregateColumns as $column) {
			$pipeline['$group']['_id'] = array_merge($pipeline['$group']['_id'], array($column => '$' . $column));
		}
		return $pipeline;
	}

    protected function processAggregate(FilterCriteria $criteria, $documents = array()) {
		if (count($aggregate = $criteria->getGroupBy()) > 0) {
            $tempCollectionName = $this->getTempCollectionName();
            /** @var \MongoCollection $tempCollection */
            $tempCollection = $this->mongoDB->getCollection($tempCollectionName);
            $tempCollection->batchInsert($documents);
			$pipeline = $this->buildAggregateExpr($aggregate);
            $result = $tempCollection->aggregate($pipeline);
            $this->mongoDB->dropCollection($tempCollectionName);
            return $result['result'];
		}
		return $documents;
    }

    /**
     * @param FilterCriteria $filterCriteria
     *
     * @return array
     */
    protected function processFilterConditions(FilterCriteria $filterCriteria) {
        $expressionBuilder = new Expr();
        foreach ($filterCriteria->getConditions() as $condition) {
            $conditionValue = Type::convertPHPToDatabaseValue($condition['value']);
            switch ($condition['operator']) {
                case FilterCriteria::EQ:
                    $expressionBuilder->field($condition['columnName'])->equals($conditionValue);
                    break;
                case FilterCriteria::IN:
                    $expressionBuilder->field($condition['columnName'])->in(Utilities::ensureIsArray($conditionValue));
                    break;
                case FilterCriteria::LT:
                    $expressionBuilder->field($condition['columnName'])->lt($conditionValue);
                    break;
                case FilterCriteria::LTE:
                    $expressionBuilder->field($condition['columnName'])->lte($conditionValue);
                    break;
                case FilterCriteria::GTE:
                    $expressionBuilder->field($condition['columnName'])->gte($conditionValue);
                    break;
                case FilterCriteria::GT:
                    $expressionBuilder->field($condition['columnName'])->gt($conditionValue);
                    break;
                case FilterCriteria::BETWEEN:
                    list($start, $end) = $conditionValue;
                    $expressionBuilder->field($condition['columnName'])->range($start, $end);
                    break;
            }
        }
        $this->processOrderBy($filterCriteria, $expressionBuilder);
        return $expressionBuilder->getQuery();
    }

    /**
     * @param                $desiredFacts
     * @param FilterCriteria $criteria
     * @param                $start
     * @param                $end
     *
     * @return array
     */
    public function findFact($desiredFacts, FilterCriteria $criteria, $start, $end) {
        $desiredFacts = Utilities::ensureIsArray($desiredFacts);
        $this->addDatesToFilter($criteria, $start, $end);
        $rows = array();
		$conditions = $this->processFilterConditions($criteria);
		foreach ($desiredFacts as $desiredFact) {
            $rows = array_merge($rows, $this->mongoDB->find($desiredFact, $conditions));
		}
        $rows = $this->processAggregate($criteria, $rows);
		return $this->prepareRowsForResultSet($rows);
    }
}